# $Id: Master.pm,v 1.11 2002/02/08 16:05:15 joern Exp $

#-----------------------------------------------------------------------
# Copyright (C) 2001-2002 Jrn Reder <joern@zyn.de> All Rights Reserved
# 
# This program is part of Video::DVDRip, which is free software; you can
# redistribute it and/or modify it under the same terms as Perl itself.
#-----------------------------------------------------------------------

package Video::DVDRip::Cluster::Master;

use base Video::DVDRip::Base;

use Event;
use constant NICE => -1;

use Video::DVDRip::Cluster::Job;
use Video::DVDRip::Cluster::Node;
use Video::DVDRip::Cluster::Project;

use Carp;
use strict;

use FileHandle;
use Data::Dumper;

sub config_filename		{ shift->{config_filename}		}
sub data_dir			{ shift->{data_dir}			}
sub node_dir			{ shift->{node_dir}			}
sub project_dir			{ shift->{project_dir}			}
sub nodes			{ shift->{nodes}			}
sub projects			{ shift->{projects}			}
sub job_id			{ shift->{job_id}			}
sub in_job_control		{ shift->{in_job_control}		}
sub node_check_watcher		{ shift->{node_check_watcher}		}

sub set_config_filename		{ shift->{config_filename}	= $_[1] }
sub set_data_dir		{ shift->{data_dir}		= $_[1] }
sub set_node_dir		{ shift->{node_dir}		= $_[1] }
sub set_project_dir		{ shift->{project_dir}		= $_[1] }
sub set_nodes			{ shift->{nodes}		= $_[1] }
sub set_projects		{ shift->{projects}		= $_[1] }
sub set_job_id			{ shift->{job_id}		= $_[1] }
sub set_in_job_control		{ shift->{in_job_control}	= $_[1] }
sub set_node_check_watcher	{ shift->{node_check_watcher}	= $_[1] }

{
	my $MASTER_OBJECT;
	
	sub get_master {
		my $class = shift;
		my %par = @_;

		# only one master object per process, so return the object
		# directly if already created.
		return $MASTER_OBJECT if $MASTER_OBJECT;

		my $self = {
			data_dir	  => $ENV{HOME}."/.dvdrip-master",
			node_dir	  => $ENV{HOME}."/.dvdrip-master/nodes",
			project_dir	  => $ENV{HOME}."/.dvdrip-master/projects",
			config_filename	  => $ENV{HOME}."/.dvdrip-master/master.conf",
			nodes		  => [],
			projects          => [],
			job_id		  => 0,
		};

		bless $self, $class;

		if ( not -d $self->data_dir ) {
			mkdir ($self->data_dir, 0755) or
				croak "can't create directory '".$self->data_dir."'";

		}

		if ( not -d $self->node_dir ) {
			mkdir ($self->node_dir, 0755) or
				croak "can't create directory '".$self->node_dir."'";

		}

		if ( not -d $self->project_dir ) {
			mkdir ($self->project_dir, 0755) or
				croak "can't create directory '".$self->project_dir."'";

		}

		$self->set_logger ( $Video::DVDRip::Cluster::logger );

		$self->log ("Master daemon activated");

		$self->load;

		$MASTER_OBJECT = $self;

		$self->enable_node_check
			if not $self->node_check_unnecessary;

		return $self;
	}

}

sub node_check_unnecessary {
	my $self = shift;
	
	if ( 0 ) {
		$self->log (
			"node check? clients=".
			Video::DVDRip::RPC::Server->instance->clients_connected.
			" projects=".@{$self->job_get_unfinished_projects}
		);
	}
	
	return if Video::DVDRip::RPC::Server->instance->clients_connected;
	return if @{$self->job_get_unfinished_projects};
	return 1;
}

sub enable_node_check {
	my $self = shift;

	return if $self->node_check_watcher;

	$self->node_check;

	my $watcher = Event->timer (
		interval => 10,
		cb       => sub { $self->node_check },
		desc     => "node check timer"
	);

	$self->log ("Node check watcher enabled");

	$self->set_node_check_watcher ( $watcher );

	1;
}

sub disable_node_check {
	my $self = shift;
	
	return if not $self->node_check_watcher;

	$self->node_check_watcher->cancel;
	$self->set_node_check_watcher(undef);

	$self->log ("Node check watcher disabled");

	1;
}

sub node_check {
	my $self = shift;
	
	my $nodes_list;
	foreach my $node ( @{$self->nodes} ) {
		$nodes_list .= $node->name." " if $node->state ne 'stopped';
	}
	
	return 1 if not $nodes_list;

	my $command = "/usr/sbin/fping $nodes_list";
	
	my $buffer;
	$self->popen (
		command      => $command,
		cb_line_read => sub {
			$buffer .= $_[0]."\n";
			1;
		},
		cb_finished => sub {
			if ( $buffer =~ /^\s*$/ ) {
				$self->log ("Warning: node check fping reported nothing");
				return;
			}
			my $node_name;
			my $idle_nodes;
			foreach my $node ( @{$self->nodes} ) {
				next if $node->state eq 'stopped';
				$node_name = $node->name;
				if ( $buffer =~ /^$node_name\s+is\s+alive/m ) {
					$self->log ("Node '$node_name' is now online")
						if not $node->alive;
					$node->set_alive(1);
				} else {
					$self->log ("Warning: Node '$node_name' is unreachable")
						if $node->alive or $node->state eq 'unknown';
					$node->set_alive(0);
				}
				++$idle_nodes if $node->state eq 'idle';
			}

			$self->disable_node_check
				if $self->node_check_unnecessary;

			$self->job_control if $idle_nodes and
					      not $self->in_job_control;
		},
	);
	
	1;
}

sub hello {
	my $self = shift;
	
	$self->enable_node_check;
	
	1;
}

sub popen {
	my $self = shift;
	my %par = @_;
	my  ($command, $cb_line_read, $cb_finished) =
	@par{'command','cb_line_read','cb_finished'};

	$self->log (5, "execute command: $command");

	my $fh = FileHandle->new;
	open ($fh, "($command) 2>&1 |") or croak "can't fork child process";

	my $line_buffer = "";

	my $waiter;
	$waiter = Event->io (
		fd   => $fh,
		poll => 'r',
		nice => NICE,
		cb   => sub {
			my ($event, $abort) = @_;
			if ( $abort or eof ($fh) ) {
				&$cb_finished if $cb_finished;
				$waiter->cancel;
				close $fh;
				$self->log (5, "command finished: $command");
				return;
			}

			my ($rc, $last_was_eol);
			while ( defined ($rc = getc($fh)) ) {
				last if $last_was_eol and $rc ne "\n" and $rc ne "\r";
				if ( $rc ne "\n" and $rc ne "\r" ) {
					$line_buffer .= $rc;
				} else {
					$last_was_eol = 1;
				}
			}

			$self->log (6, "2 got data: ".$line_buffer);
			# !!! STRANGE BUG WORKAROUND !!!
			# otherwise $event appears at $_[0] in
			# (some, not all) called closures !!?!
			$_[0] = $line_buffer;
			&$cb_line_read($line_buffer) if &$cb_line_read;
			$line_buffer = $rc;

			1;
				
		},
		desc => "command execution",
	);
	
	return $waiter;
}

sub load {
	my $self = shift;
	
	my $filename = $self->config_filename;
	return if not -f $filename;
	
	my $fh = FileHandle->new;
	open ($fh, $filename)
		or croak "can't read master config file '$filename'";
	my $data_blob = join ('', <$fh>);
	close $fh;
	
	my $data;
	$data = eval $data_blob;
	croak "Error loading master config file '$filename': $@" if $@;
	
	$self->set_job_id ( $data->{job_id} );

	$self->load_nodes;
	$self->load_projects ( project_order => $data->{project_order} );
	
	1;
}

sub save {
	my $self = shift;
	
	my $filename = $self->config_filename;
	
	my @project_order = map { $_->filename } @{$self->projects};
	
	my $data = {
		job_id        => $self->job_id,
		project_order => \@project_order,
	};

	my $dd = Data::Dumper->new ( [$data], ['data'] );
	$dd->Indent(1);
	my $data_blob = $dd->Dump;
	
	my $fh = FileHandle->new;
	open ($fh, "> $filename")
		or croak "can't write master config file '$filename'";
	print $fh $data_blob;
	close $fh;
	
	1;
}

sub load_nodes {
	my $self = shift;
	
	my $dir = $self->node_dir;
	
	my @nodes;
	foreach my $file ( <$dir/*> ) {
		$self->log ("Loading node file '$file'");
		my $node = Video::DVDRip::Cluster::Node->new_from_file (
			filename  => $file,
		);
		$node->reset;
		push @nodes, $node;
	}
	
	$self->set_nodes( \@nodes );
	
	1;
}

sub load_projects {
	my $self = shift;
	my %par = @_;
	my ($project_order) = @par{'project_order'};

	my $dir = $self->project_dir;

	my @projects;	
	foreach my $filename ( @{$project_order} ) {
		next if not -r $filename;
		$self->log ("Loading project file '$filename'");
		my $project = Video::DVDRip::Cluster::Project->new_from_file (
			filename  => $filename,
		);
		$project->reset;
		push @projects, $project;
	}

	$self->set_projects (\@projects);
	
	1;
}

sub add_node {
	my $self = shift;
	my %par = @_;
	my ($node) = @par{'node'};
	
	my $filename = $self->node_dir.'/'.$node->name.'.node';

	croak "msg: Node must have a name" if $node->name eq '';
	croak "msg: Node with this name already exists" if -f $filename;
	
	push @{$self->nodes}, $node;
	$node->set_filename ($filename);
	$node->save;
	
	$self->log ("Node '".$node->name."' saved to '$filename'");
	
	1;
}

sub remove_node {
	my $self = shift;
	my %par = @_;
	my ($node) = @par{'node'};

	my $i=0;
	foreach my $n ( @{$self->nodes} ) {
		last if $n == $node;
		++$i;
	}
	
	croak "Unknown node $node" if $i == @{$self->nodes};
	
	unlink $node->filename;

	splice @{$self->nodes}, $i, 1;

	1;
}

sub get_project_index {
	my $self = shift;
	my %par = @_;
	my ($project) = @par{'project'};
	
	my $projects = $self->projects;
	
	my $i = 0;
	foreach my $p ( @{$projects} ) {
		last if $p == $project;
		++$i;
	}
	
	croak "Unknown project $project" if $i == @{$projects};
	
	return $i;
}

sub add_project {
	my $self = shift;
	my %par = @_;
	my ($project) = @par{'project'};
	
	push @{$self->projects}, $project;
	
	my $job_id = $self->set_job_id ( 1 + $self->job_id );
	my $filename = sprintf (
		"%s/%08d-%s.rip",
		$self->project_dir,
		$job_id,
		$project->name
	);

	$project->set_filename ($filename);
	$project->set_state ('not scheduled');
	$project->set_id ($job_id);

	# calculate chunk_cnt for this title
	my $title = $project->title;
	my $chunk_cnt = int($title->frames / 10000) ||
			@{$self->nodes};

	$self->log (
		"Project has ".$title->frames." frames. ".
		"Resulting chunk count: $chunk_cnt"
	);

	$title->set_chunk_cnt ( $chunk_cnt );

	# save changes to project
	$project->save;

	$self->log ("Project with filename '$filename' added");
	
	# save new state
	$self->save;

	1;
}

sub move_up_project {
	my $self = shift;
	my %par = @_;
	my ($project) = @par{'project'};
	
	my $i = $self->get_project_index ( project => $project );

	# already on top?	
	return if $i == 0;
	
	# move project up
	my $projects = $self->projects;
	@{$projects}[$i,$i-1] = @{$projects}[$i-1,$i];

	# save new state
	$self->save;

	1;
}

sub move_down_project {
	my $self = shift;
	my %par = @_;
	my ($project) = @par{'project'};
	
	my $i = $self->get_project_index ( project => $project );
	
	# already on bottom?
	my $projects = $self->projects;
	return if $i == @{$projects} - 1;
	
	# move project up
	@{$projects}[$i,$i+1] = @{$projects}[$i+1,$i];
	
	# save new state
	$self->save;

	1;
}

sub schedule_project {
	my $self = shift;
	my %par = @_;
	my ($project) = @par{'project'};
	
	# check for existence
	$self->get_project_index ( project => $project );

	# change project state
	$project->set_state ( 'waiting' );
	
	# save project's state
	$project->save;

	# maybe the job controller can dispose some work now...
	$self->job_control;

	1;
}

sub remove_project {
	my $self = shift;
	my %par = @_;
	my ($project) = @par{'project'};
	
	# check for existence
	my $i = $self->get_project_index ( project => $project );

	# check project state
	return 1 if $project->state ne 'waiting' and
		    $project->state ne 'not scheduled' and
		    $project->state ne 'merged' and
		    $project->state ne 'audio processed' and
		    $project->state ne 'avi splitted' and
	            $project->state ne 'finished';

	unlink $project->filename;
	splice @{$self->projects}, $i, 1;

	$self->log ("Project ".$project->label." removed");

	1;
}

sub job_control {
	my $self = shift;
	return if $self->in_job_control;

	$self->set_in_job_control(1);

	# 1. move waiting projects before not scheduled projects
	$self->job_sort_projects;

	# 2. delegate pending jobs
	$self->job_delegate_pending_jobs;

	$self->set_in_job_control(0);

	1;
}

sub job_sort_projects {
	my $self = shift;
	
	# divide projects in order to not scheduled and scheduled projects
	my @not_scheduled;
	my @scheduled;
	my @finished;
	foreach my $project ( @{$self->projects} ) {
		if ( $project->state eq 'not scheduled' ) {
			push @not_scheduled, $project;
		} elsif ( $project->state eq 'finished' ) {
			push @finished, $project;
		} else {
			push @scheduled, $project;
		}
	}
	
	# define project list with scheduled first, then the not
	# scheduled projects
	my @projects = ( @scheduled, @not_scheduled, @finished );
	
	$self->set_projects ( \@projects );

	1;
}

sub job_delegate_pending_jobs {
	my $self = shift;

	# get idle nodes
	my $idle_nodes = $self->job_get_idle_nodes;
	
	# nothing to do if no node is idle
	return 1 if not @{$idle_nodes};	

	# get waiting projects
	my $unfinished_projects = $self->job_get_unfinished_projects;

	# nothing to do if no project is in the queue for processing
	return 1 if not @{$unfinished_projects};	

#	$self->log ("Job control: we have ".@{$idle_nodes}." idle node(s)");
#	$self->log ("Job control: we have ".@{$unfinished_projects}." unfinished project(s)");

	# Ok: we have some idle nodes and unfinished projects.
	#     That means: we can do some work!
	
	# we try to get projects finished, so we iterate
	# over all project canditates and dispose nodes
	# to them, as long as we have idle nodes and
	# unfinished projects.
	foreach my $project ( @{$unfinished_projects} ) {
		last if not @{$idle_nodes};

		# is transcoding or merging of this projet finished?
		if ( $project->state eq 'transcoded' or
		     $project->state eq 'merged' or
		     $project->state eq 'audio processed' ) {

			# we prefer the node with local harddisk access
			# for these jobs
			my $node;
			for ( my $i=0; $i < @{$idle_nodes}; ++$i ) {
				if ( $idle_nodes->[$i]->data_is_local ) {
					$node = splice @{$idle_nodes}, $i, 1;
					last;
				}
			}

			# otherwise take the next free node
			$node ||= shift @{$idle_nodes};
			
			# and start the merge, audio or avisplit job on it
			if ( $project->state eq 'transcoded' ) {
				Video::DVDRip::Cluster::Job->new (
					project => $project,
					node    => $node,
				)->start_merging;

			} elsif ( $project->state eq 'merged' ) {
				Video::DVDRip::Cluster::Job->new (
					project => $project,
					node    => $node,
				)->start_audio_processing;

			} else {
				Video::DVDRip::Cluster::Job->new (
					project => $project,
					node    => $node,
				)->start_avisplit;

			}
		}
	
		# otherwise start transcode jobs as long as we have
		# free nodes and free chunks in this project
		my $chunk_nr;
		while ( @{$idle_nodes} and
			defined ($chunk_nr = $project->title->get_next_chunk_nr) ) {

			# start transcode job
			Video::DVDRip::Cluster::Job->new (
				project => $project,
				node    => shift @{$idle_nodes},
			)->start_transcoding ( chunk_nr => $chunk_nr );
		}
	}

	1;
}

sub job_get_idle_nodes {
	my $self = shift;
	
	# create a list of nodes which have the 'idle' state
	my @idle_nodes;
	foreach my $node ( @{$self->nodes} ) {
		push @idle_nodes, $node if $node->state eq 'idle';
	}
	
	@idle_nodes = sort { $b->speed <=> $a->speed } @idle_nodes;

	return \@idle_nodes;
}

sub job_get_unfinished_projects {
	my $self = shift;
	
	# create a list of nodes which are not finished yet
	# (waiting or runnding, but not all chunks processed)
	my @wait_for_transcode_projects;
	my @wait_for_merge_projects;
	my @wait_for_audio_projects;
	my @wait_for_avisplit_projects;

	foreach my $project ( @{$self->projects} ) {
		push @wait_for_transcode_projects, $project
			if $project->state eq 'waiting' or
			   ($project->state eq 'transcoding' and
			    $project->title->chunks_finished <
			    $project->title->chunk_cnt);

		push @wait_for_merge_projects, $project
			if ( $project->state eq 'transcoded' );

		push @wait_for_audio_projects, $project
			if ( $project->state eq 'merged' );

		push @wait_for_avisplit_projects, $project
			if ( $project->state eq 'audio processed' );
	}
	
	# wait for audio / merge projects have a higher priority,
	# because this is the last step before the project
	# is finished.
	my @projects = ( @wait_for_avisplit_projects,
			 @wait_for_audio_projects,
			 @wait_for_merge_projects,
			 @wait_for_transcode_projects );

	$self->enable_node_check if @projects;

	return \@projects;
}

sub shutdown {
	my $self = shift;
	
	Event->timer (
		interval => 2,
		cb       => sub { Event::unloop_all() },
		desc     => "dvd::rip shutdown timer"
	);

	$self->log ("Cluster control daemon will shutdown in 2 seconds...");

	1;
	
}

1;
